/* xlbfun.c - xlisp basic builtin functions */
#include "xlset.h"
#include "xlbfun.h"
#include "xldmem.h"
#include "xleval.h"
#include "xlsubr.h"
#include "xlsym.h"
#include "xlinit.h"



/* local variables */
static char gsprefix[STRMAX+1] = { 'G', 0 };
static char gsnumber = 1;


/* xeval - the builtin function 'eval' */
NODE *xeval(args)
NODE *args;
{
	NODE *oldstk,expr,*val;

	/* create a new stack frame */
	oldstk = xlsave(&expr,NULL);

	/* get the expression to evaluate */
	expr.n_ptr = xlarg(&args);
	xllastarg(args);

	/* evaluate the expression */
	val = xleval(expr.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the expression evaluated */
	return (val);
}

/* xapply - the builtin function 'apply' */
NODE *xapply(args)
NODE *args;
{
	NODE *oldstk,fun,arglist,*val;

	/* create a new stack frame */
	oldstk = xlsave(&fun,&arglist,NULL);

	/* get the function and argument list */
	fun.n_ptr = xlarg(&args);
	arglist.n_ptr = xlarg(&args);
	xllastarg(args);

	/* if the function is a symbol, get its value */
	if (fun.n_ptr && fun.n_ptr->n_type == SYM) fun.n_ptr = xleval(fun.n_ptr);

	/* apply the function to the arguments */
	val = xlapply(fun.n_ptr,arglist.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the expression evaluated */
	return (val);
}

/* xfuncall - the builtin function 'funcall' */
NODE *xfuncall(args)
NODE *args;
{
	NODE *oldstk,fun,arglist,*val;

	/* create a new stack frame */
	oldstk = xlsave(&fun, &arglist, NULL);

	/* get the function and argument list */
	fun.n_ptr = xlarg(&args);
	arglist.n_ptr = args;

	/* if the function is a symbol, get its value */
	if (fun.n_ptr && fun.n_ptr->n_type == SYM) fun.n_ptr = xleval(fun.n_ptr);

	/* apply the function to the arguments */
	val = xlapply(fun.n_ptr, arglist.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the expression evaluated */
	return (val);
}

/* xquote - builtin function to quote an expression */
NODE *xquote(args)
NODE *args;
{
	/* make sure there is exactly one argument */
	if (args == NULL || args->n_listnext != NULL) xlfail("incorrect number of arguments");

	/* return the quoted expression */
	return (args->n_listvalue);
}

/* xset - builtin function set */
NODE *xset(args)
NODE *args;
{
	NODE *sym,*val;

	/* get the symbol and new value */
	sym = xlmatch(SYM, &args);
	val = xlarg(&args);
	xllastarg(args);

	/* assign the symbol the value of argument 2 and the return value */
	assign(sym, val);

	/* return the result value */
	return (val);
}

/* xsetq - builtin function setq */
NODE *xsetq(args)
NODE *args;
{
	NODE *oldstk,arg,sym,val;

	/* create a new stack frame */
	oldstk = xlsave(&arg, &sym, &val, NULL);

	/* initialize */
	arg.n_ptr = args;

	/* get the symbol and new value */
	sym.n_ptr = xlmatch(SYM, &arg.n_ptr);
	val.n_ptr = xlevarg(&arg.n_ptr);
	xllastarg(arg.n_ptr);

	/* assign the symbol the value of argument 2 and the return value */
	assign(sym.n_ptr, val.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the result value */
	return (val.n_ptr);
}

/* defun - internal function definition routine */
LOCAL NODE *defun(args, type)
NODE *args;
NODE *type;
{
	NODE *oldstk, sym, fargs, fun;

	/* create a new stack frame */
	oldstk = xlsave(&sym, &fargs, &fun, NULL);

	/* get the function symbol and formal argument list */
	sym.n_ptr = xlmatch(SYM, &args);
	fargs.n_ptr = xlmatch(LIST, &args);

	/* create a new function definition */
	fun.n_ptr = newnode(LIST);
	fun.n_ptr->n_listvalue = type;
	fun.n_ptr->n_listnext = newnode(LIST);
	fun.n_ptr->n_listnext->n_listvalue = fargs.n_ptr;
	fun.n_ptr->n_listnext->n_listnext = args;

	/* make the symbol point to a new function definition */
	assign(sym.n_ptr, fun.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the function symbol */
	return (sym.n_ptr);
}

/* xdefun - builtin function 'defun' */
NODE *xdefun(args)
NODE *args;
{
	return (defun(args, s_lambda));
}

/* xndefun - builtin function 'ndefun' */
NODE *xndefun(args)
NODE *args;
{
	return (defun(args, s_nlambda));
}


/* xgensym - generate a symbol */
NODE *xgensym(args)
NODE *args;
{
	char sym[STRMAX+1];
	NODE *x;

	/* get the prefix or number */
	if (args) {
		if (NULL != (x = xlarg(&args))) {
			switch (x->n_type) {
				case SYM:
					strcpy(gsprefix, xlsymname(x));
					break;
				case STR:
					strcpy(gsprefix, x->n_str);
					break;
				case INT:
					gsnumber = x->n_int;
					break;
				default:
					xlfail("bad argument type");
			}
		}
	}
	xllastarg(args);

	/* create the pname of the new symbol */
	sprintf(sym, "%s%d", gsprefix, gsnumber++);

	/* make a symbol with this print name */
	return (xlmakesym(sym, DYNAMIC));
}

/* xintern - intern a symbol */
NODE *xintern(args)
NODE *args;
{
	NODE *oldstk, sym;

	/* create a new stack frame */
	oldstk = xlsave(&sym, NULL);

	/* get the symbol to intern */
	sym.n_ptr = xlmatch(SYM, &args);
	xllastarg(args);

	/* intern the symbol */
	sym.n_ptr = xlintern(sym.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the symbol */
	return (sym.n_ptr);
}

/* xsymname - get the print name of a symbol */
NODE *xsymname(args)
NODE *args;
{
	NODE *sym;

	/* get the symbol */
	sym = xlmatch(SYM, &args);
	xllastarg(args);

	/* return the print name */
	return (sym->n_symplist->n_listvalue);
}

/* xsymplist - get the property list of a symbol */
NODE *xsymplist(args)
NODE *args;
{
	NODE *sym;

	/* get the symbol */
	sym = xlmatch(SYM, &args);
	xllastarg(args);

	/* return the property list */
	return (sym->n_symplist->n_listnext);
}

/* xget - get the value of a property */
NODE *xget(args)
NODE *args;
{
	NODE *sym, *prp;

	/* get the symbol and property */
	sym = xlmatch(SYM, &args);
	prp = xlmatch(SYM, &args);
	xllastarg(args);

	/* retrieve the property value */
	return (xlgetprop(sym, prp));
}

/* xputprop - put a property value onto a property list */
NODE *xputprop(args)
NODE *args;
{
	NODE *oldstk, sym, val, prp;

	/* create a new stack frame */
	oldstk = xlsave(&sym, &val, &prp, NULL);

	/* get the symbol, value and property */
	sym.n_ptr = xlmatch(SYM, &args);
	val.n_ptr = xlarg(&args);
	prp.n_ptr = xlmatch(SYM, &args);
	xllastarg(args);

	/* put the property onto the property list */
	xlputprop(sym.n_ptr, val.n_ptr, prp.n_ptr);

	/* restore the previouse stack frame */
	xlstack = oldstk;

	/* return the value */
	return (val.n_ptr);
}

/* xremprop - remove a property value from a property list */
NODE *xremprop(args)
NODE *args;
{
	NODE *sym, *prp;

	/* get the symbol and property */
	sym = xlmatch(SYM, &args);
	prp = xlmatch(SYM, &args);
	xllastarg(args);

	/* remove the property */
	xlremprop(sym, prp);

	/* return nil */
	return (NULL);
}
