/* xleval - xlisp evaluator */
#include "xlset.h"
#include "xleval.h"
#include "xlbind.h"
#include "xldmem.h"
#include "xlio.h"
#include "xlobj.h"
#include "xlprin.h"
#include "xlsubr.h"
#include "xlisp.h"
#include "xlsubr.h"
#include "xlsym.h"
#include "xlsys.h"
#include "xlinit.h"

/* global variables */
NODE *xlstack;

/* trace stack */
static NODE *trace_stack[TDEPTH];
static int trace_pointer;

/* xlsave - save nodes on the stack */
NODE *xlsave(NODE* n, ...)
/* NODE *n; */
{
	NODE **nptr, *oldstk;

	/* save the old stack pointer */
	oldstk = xlstack;

	/* save each node */
	for (nptr = &n; *nptr != NULL; nptr++) {
		(*nptr)->n_type = LIST;
		(*nptr)->n_listvalue = NULL;
		(*nptr)->n_listnext = xlstack;
		xlstack = *nptr;
	}

	/* return the old stack pointer */
	return (oldstk);
}

/* tpush - add an entry to the trace stack */
LOCAL void tpush(nptr)
NODE *nptr;
{
	if (++trace_pointer < TDEPTH) trace_stack[trace_pointer] = nptr;
}

/* tpop - pop an entry from the trace stack */
LOCAL void tpop()
{
	trace_pointer--;
}

/* iskeyword - check to see if a symbol is a keyword */
LOCAL int iskeyword(sym)
NODE *sym;
{
	return (sym == k_rest || sym == k_aux);
}

/* xlabind - bind the arguments for a function */
void xlabind(fargs, aargs)
NODE *fargs;
NODE *aargs;
{
	NODE *oldstk, farg, aarg, *arg;

	/* create a stack frame */
	oldstk = xlsave(&farg, &aarg, NULL);

	/* initialize the pointers */
	farg.n_ptr = fargs;
	aarg.n_ptr = aargs;

	/* evaluate and bind each argument */
	while (farg.n_ptr != NULL && aarg.n_ptr != NULL) {

		/* check for a keyword */
		if (iskeyword(arg = farg.n_ptr->n_listvalue)) break;

		/* bind the formal variable to the argument value */
		xlbind(arg, aarg.n_ptr->n_listvalue);

		/* move the argument list pointers ahead */
		farg.n_ptr = farg.n_ptr->n_listnext;
		aarg.n_ptr = aarg.n_ptr->n_listnext;
	}

	/* check for the '&rest' keyword */
	if (farg.n_ptr && farg.n_ptr->n_listvalue == k_rest) {
		farg.n_ptr = farg.n_ptr->n_listnext;
		if (farg.n_ptr && (arg = farg.n_ptr->n_listvalue) && !iskeyword(arg)) {
			xlbind(arg,aarg.n_ptr);
		} else {
			xlfail("symbol missing after &rest");
		}
		farg.n_ptr = farg.n_ptr->n_listnext;
		aarg.n_ptr = NULL;
	}

	/* check for the '&aux' keyword */
	if (farg.n_ptr && farg.n_ptr->n_listvalue == k_aux) {
		while ((farg.n_ptr = farg.n_ptr->n_listnext) != NULL) xlbind(farg.n_ptr->n_listvalue,NULL);
	}

	/* make sure the correct number of arguments were supplied */
	if (farg.n_ptr != aarg.n_ptr) xlfail("incorrect number of arguments to a function");

	/* restore the previous stack frame */
	xlstack = oldstk;
}

/* evfun - evaluate a function */
LOCAL NODE *evfun(fun, args)
NODE *fun;
NODE *args;
{
	NODE *oldenv, *oldstk, cptr, *fargs, *val = NULL;

	/* create a stack frame */
	oldstk = xlsave(&cptr, NULL);

	/* skip the function type */
	if ((fun = fun->n_listnext) == NULL) xlfail("bad function definition");

	/* get the formal argument list */
	if ((fargs = fun->n_listvalue) != NULL && fargs->n_type != LIST) xlfail("bad formal argument list");

	/* bind the formal parameters */
	oldenv = xlenv;
	xlabind(fargs, args);
	xlfixbindings(oldenv);

	/* execute the code */
	for (cptr.n_ptr = fun->n_listnext; cptr.n_ptr != NULL; ) val = xlevarg(&cptr.n_ptr);

	/* restore the environment */
	xlunbind(oldenv);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the result value */
	return (val);
}

/* evform - evaluate a form */
LOCAL NODE *evform(nptr)
NODE *nptr;
{
	NODE *oldstk, fun, args, *val = NULL, *type;

	/* create a stack frame */
	oldstk = xlsave(&fun, &args, NULL);

	/* get the function and the argument list */
	fun.n_ptr = nptr->n_listvalue;
	args.n_ptr = nptr->n_listnext;

	/* evaluate the first expression */
	if ((fun.n_ptr = xleval(fun.n_ptr)) == NULL) xlfail("null function");

	/* evaluate the function */
	switch (fun.n_ptr->n_type) {
		case SUBR:
			args.n_ptr = xlevlist(args.n_ptr);
		case FSUBR:
			val = (*fun.n_ptr->n_subr)(args.n_ptr);
			break;
		case LIST:
			if ((type = fun.n_ptr->n_listvalue) == s_lambda) {
				args.n_ptr = xlevlist(args.n_ptr);
			} else if (type != s_nlambda) {
				xlfail("bad function type");
			}
			val = evfun(fun.n_ptr,args.n_ptr);
			break;
		case OBJ:
			val = xlsend(fun.n_ptr,args.n_ptr);
			break;
		default:
			xlfail("bad function");
	}

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the result value */
	return (val);
}

/* evsym - evaluate a symbol */
LOCAL NODE *evsym(sym)
NODE *sym;
{
	NODE *p;

	/* check for a current object */
	if ((p = xlobsym(sym)) != NULL) {
		return (p->n_listvalue);
	} else if ((p = sym->n_symvalue) == s_unbound) {
		xlfail("unbound variable");
	}
	return (p);
}

/* xleval - evaluate an xlisp expression */
NODE *xleval(expr)
NODE *expr;
{
	/* evaluate null to itself */
	if (expr == NULL) return (NULL);

	/* add trace entry */
	tpush(expr);

	/* check type of value */
	switch (expr->n_type) {
		case LIST:
			expr = evform(expr);
			break;
		case SYM:
			expr = evsym(expr);
			break;
		case INT:
		case STR:
		case SUBR:
		case FSUBR:
			break;
		default:
			xlfail("can't evaluate expression");
	}

	/* remove trace entry */
	tpop();

	/* return the value */
	return (expr);
}

/* xlapply - apply a function to a list of arguments */
NODE *xlapply(fun, args)
NODE *fun;
NODE *args;
{
	NODE *val = NULL;

	/* check for a null function */
	if (fun == NULL) xlfail("null function");

	/* evaluate the function */
	switch (fun->n_type) {
		case SUBR:
			val = (*fun->n_subr)(args);
			break;
		case LIST:
			if (fun->n_listvalue != s_lambda) xlfail("bad function type");
			val = evfun(fun, args);
			break;
		default:
			xlfail("bad function");
	}

	/* return the result value */
	return (val);
}

/* xlevlist - evaluate a list of arguments */
NODE *xlevlist(args)
NODE *args;
{
	NODE *oldstk, src, dst, *new, *last = NULL, *val;

	/* create a stack frame */
	oldstk = xlsave(&src, &dst, NULL);

	/* initialize */
	src.n_ptr = args;

	/* evaluate each argument */
	for (val = NULL; src.n_ptr; src.n_ptr = src.n_ptr->n_listnext) {

		/* check this entry */
		if (src.n_ptr->n_type != LIST) xlfail("bad argument list");

		/* allocate a new list entry */
		new = newnode(LIST);
		if (val) {
			last->n_listnext = new;
		} else {
			val = dst.n_ptr = new;
		}
		new->n_listvalue = xleval(src.n_ptr->n_listvalue);
		last = new;
	}

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the new list */
	return (val);
}

/* stdprint - print to standard output */
void stdprint(expr)
NODE *expr;
{
	xlprint(s_stdout->n_symvalue, expr, TRUE);
	xlterpri(s_stdout->n_symvalue);
}

/* baktrace - do a back trace */
LOCAL void baktrace()
{
	for (; trace_pointer >= 0; trace_pointer--) if (trace_pointer < TDEPTH) stdprint(trace_stack[trace_pointer]);
}

/* xlfail - error handling routine */
void xlfail(err)
char *err;
{
	/* print the error message */
	printf("error: %s\n", err);

	/* flush the terminal input buffer */
	xlflush();

	/* unbind bound symbols */
	xlunbind(NULL);

	/* do the back trace */
	if (s_tracenable->n_symvalue) baktrace();
	trace_pointer = -1;

	/* restart */
	longjmp(*xljmpbuf, 1); /* was longjmp(xljmpbuf, 1); */
}

/* xleinit - initialize the evaluator */
void xleinit(void)
{
	/* initialize debugging stuff */
	trace_pointer = -1;
}
