/* xllist - xlisp list builtin functions */
#include "xlset.h"
#include "xllist.h"
#include "xldmem.h"
#include "xleval.h"
#include "xlisp.h"
#include "xllist.h"
#include "xlmath.h"
#include "xlstr.h"
#include "xlsubr.h"
#include "xlsym.h"
#include "xlsys.h"
#include "xlinit.h"

/* xcar - return the car of a list */
NODE *xcar(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its car */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	return (list ? list->n_listvalue : ((NODE *) NULL));
}

/* xcaar - return the caar of a list */
NODE *xcaar(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its caar */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	if (list) list = list->n_listvalue;
	return (list ? list->n_listvalue : ((NODE *) NULL));
}

/* xcadr - return the cadr of a list */
NODE *xcadr(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its cadr */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	if (list) list = list->n_listnext;
	return (list ? list->n_listvalue : ((NODE *) NULL));
}

/* xcdr - return the cdr of a list */
NODE *xcdr(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its cdr */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	return (list ? list->n_listnext : ((NODE *) NULL));
}

/* xcdar - return the cdar of a list */
NODE *xcdar(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its cdar */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	if (list) list = list->n_listvalue;
	return (list ? list->n_listnext : ((NODE *) NULL));
}

/* xcddr - return the cddr of a list */
NODE *xcddr(args)
NODE *args;
{
	NODE *list;

	/* get the list and return its cddr */
	list = xlmatch(LIST, &args);
	xllastarg(args);
	if (list) list = list->n_listnext;
	return (list ? list->n_listnext : ((NODE *) NULL));
}

/* xcons - construct a new list cell */
NODE *xcons(args)
NODE *args;
{
	NODE *arg1, *arg2, *val;

	/* get the two arguments */
	arg1 = xlarg(&args);
	arg2 = xlarg(&args);
	xllastarg(args);

	/* construct a new list element */
	val = newnode(LIST);
	val->n_listvalue = arg1;
	val->n_listnext  = arg2;

	/* return the list */
	return (val);
}

/* xlist - built a list of the arguments */
NODE *xlist(args)
NODE *args;
{
	NODE *oldstk, arg, list, val, *last, *lptr;

	/* create a new stack frame */
	oldstk = xlsave(&arg, &list, &val, NULL);

	/* initialize */
	arg.n_ptr = args;

	/* evaluate and append each argument */
	for (last = NULL; arg.n_ptr != NULL; last = lptr) {

		/* evaluate the next argument */
		val.n_ptr = xlarg(&arg.n_ptr);

		/* append this argument to the end of the list */
		lptr = newnode(LIST);
		if (last == NULL) {
			list.n_ptr = lptr;
		} else last->n_listnext = lptr;

		lptr->n_listvalue = val.n_ptr;
	}

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the list */
	return (list.n_ptr);
}

/* xappend - builtin function append */
NODE *xappend(args)
NODE *args;
{
	NODE *oldstk, arg, list, last, val, *lptr;

	/* create a new stack frame */
	oldstk = xlsave(&arg, &list, &last, &val, NULL);

	/* initialize */
	arg.n_ptr = args;

	/* evaluate and append each argument */
	while (arg.n_ptr != NULL) {

		/* evaluate the next argument */
		list.n_ptr = xlmatch(LIST, &arg.n_ptr);

		/* append each element of this list to the result list */
		while (list.n_ptr != NULL && list.n_ptr->n_type == LIST) {

			/* append this element */
			lptr = newnode(LIST);
			if (last.n_ptr == NULL) {
				val.n_ptr = lptr;
			} else last.n_ptr->n_listnext = lptr;

			lptr->n_listvalue = list.n_ptr->n_listvalue;

			/* save the new last element */
			last.n_ptr = lptr;

			/* move to the next element */
			list.n_ptr = list.n_ptr->n_listnext;
		}

		/* make sure the list ended in a nil */
		if (list.n_ptr != NULL) xlfail("bad list");
	}

	/* restore previous stack frame */
	xlstack = oldstk;

	/* return the list */
	return (val.n_ptr);
}

/* xreverse - builtin function reverse */
NODE *xreverse(args)
NODE *args;
{
	NODE *oldstk, list, val, *lptr;

	/* create a new stack frame */
	oldstk = xlsave(&list, &val, NULL);

	/* get the list to reverse */
	list.n_ptr = xlmatch(LIST, &args);
	xllastarg(args);

	/* append each element of this list to the result list */
	while (list.n_ptr != NULL && list.n_ptr->n_type == LIST) {

		/* append this element */
		lptr = newnode(LIST);
		lptr->n_listvalue = list.n_ptr->n_listvalue;
		lptr->n_listnext = val.n_ptr;
		val.n_ptr = lptr;

		/* move to the next element */
		list.n_ptr = list.n_ptr->n_listnext;
	}

	/* make sure the list ended in a nil */
	if (list.n_ptr != NULL) xlfail("bad list");

	/* restore previous stack frame */
	xlstack = oldstk;

	/* return the list */
	return (val.n_ptr);
}

/* xlast - return the last cons of a list */
NODE *xlast(args)
NODE *args;
{
	NODE *list;

	/* get the list */
	list = xlmatch(LIST, &args);
	xllastarg(args);

	/* find the last cons */
	while (list && list->n_type == LIST && list->n_listnext) list = list->n_listnext;

	/* make sure the list ended correctly */
	if (list == NULL && list->n_type != LIST) xlfail("bad list");

	/* return the last element */
	return (list);
}

/* member - internal member function */
LOCAL NODE *member(args,fcn)
NODE *args;
int (*fcn)();
{
	NODE *x, *list;

	/* get the expression to look for and the list */
	x = xlarg(&args);
	list = xlmatch(LIST, &args);
	xllastarg(args);

	/* look for the expression */
	for (; list && list->n_type == LIST; list = list->n_listnext) {
		if ((*fcn)(x,list->n_listvalue)) return (list);
	}

	/* return failure indication */
	return (NULL);
}

/* eq - internal eq function */
LOCAL int eq(arg1, arg2)
NODE *arg1;
NODE *arg2;
{
	/* compare the arguments */
	if (arg1 != NULL && arg1->n_type == INT && arg2 != NULL && arg2->n_type == INT) return (arg1->n_int == arg2->n_int);

	return (arg1 == arg2);
}

/* equal - internal equal function */
LOCAL int equal(arg1,arg2)
NODE *arg1;
NODE *arg2;
{
	/* compare the arguments */
	if (eq(arg1, arg2)) {
		return (TRUE);
	} else if (arg1 && arg1->n_type == LIST &&  arg2 && arg2->n_type == LIST) {
		return (equal(arg1->n_listvalue,arg2->n_listvalue) && equal(arg1->n_listnext, arg2->n_listnext));
	}
	return (FALSE);
}

/* xmember - builtin function 'member' */
NODE *xmember(args)
NODE *args;
{
	return (member(args, equal));
}

/* xmemq - builtin function 'memq' */
NODE *xmemq(args)
NODE *args;
{
	return (member(args, eq));
}

/* afind - find a pair in an association list */
LOCAL NODE *afind(expr, alist, fcn)
NODE *expr, *alist;
int (*fcn)();
{
	NODE *pair;

	for (; alist && alist->n_type == LIST; alist = alist->n_listnext) {
		if ((pair = alist->n_listvalue) && pair->n_type == LIST) {
			if ((*fcn)(expr, pair->n_listvalue)) {
				return (pair);
			}
		}
	}
	return (NULL);
}

/* assoc - internal assoc function */
LOCAL NODE *assoc(args, fcn)
NODE *args;
int (*fcn)();
{
	NODE *expr,*alist;

	/* get the expression to look for and the association list */
	expr = xlarg(&args);
	alist = xlmatch(LIST, &args);
	xllastarg(args);

	/* look for the expression */
	return (afind(expr, alist, fcn));
}

/* xassoc - builtin function 'assoc' */
NODE *xassoc(args)
NODE *args;
{
	return (assoc(args, equal));
}

/* xassq - builtin function 'assq' */
NODE *xassq(args)
NODE *args;
{
	return (assoc(args, eq));
}

/* subst - substitute one expression for another */
LOCAL NODE *subst(to, from, expr)
NODE *to;
NODE *from;
NODE *expr;
{
	NODE *oldstk, car, cdr, *val;

	if (eq(expr, from))
		val = to;
	else if (expr == NULL || expr->n_type != LIST)
		val = expr;
	else {
		oldstk = xlsave(&car, &cdr, NULL);
		car.n_ptr = subst(to, from, expr->n_listvalue);
		cdr.n_ptr = subst(to, from, expr->n_listnext);
		val = newnode(LIST);
		val->n_listvalue = car.n_ptr;
		val->n_listnext = cdr.n_ptr;
		xlstack = oldstk;
	}
	return (val);
}

/* xsubst - substitute one expression for another */
NODE *xsubst(args)
NODE *args;
{
	NODE *oldstk, to, from, expr, *val;

	/* create a new stack frame */
	oldstk = xlsave(&to, &from, &expr, NULL);

	/* get the to value, the from value and the expression */
	to.n_ptr = xlarg(&args);
	from.n_ptr = xlarg(&args);
	expr.n_ptr = xlarg(&args);
	xllastarg(args);

	/* do the substitution */
	val = subst(to.n_ptr, from.n_ptr, expr.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the result */
	return (val);
}

/* sublis - substitute using an association list */
LOCAL NODE *sublis(alist, expr)
NODE *alist;
NODE *expr;
{
	NODE *oldstk, car, cdr, *val;

	if ((val = afind(expr, alist, eq)))
		val = val->n_listnext;
	else if (expr == NULL || expr->n_type != LIST)
		val = expr;
	else {
		oldstk = xlsave(&car, &cdr, NULL);
		car.n_ptr = sublis(alist, expr->n_listvalue);
		cdr.n_ptr = sublis(alist, expr->n_listnext);
		val = newnode(LIST);
		val->n_listvalue = car.n_ptr;
		val->n_listnext = cdr.n_ptr;
		xlstack = oldstk;
	}
	return (val);
}

/* xsublis - substitute using an association list */
NODE *xsublis(args)
NODE *args;
{
	NODE *oldstk, alist, expr, *val;

	/* create a new stack frame */
	oldstk = xlsave(&alist, &expr, NULL);

	/* get the assocation list and the expression */
	alist.n_ptr = xlmatch(LIST, &args);
	expr.n_ptr = xlarg(&args);
	xllastarg(args);

	/* do the substitution */
	val = sublis(alist.n_ptr, expr.n_ptr);

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the result */
	return (val);
}

/* nth - internal nth function */
LOCAL NODE *nth(args, cdrflag)
NODE *args;
int cdrflag;
{
	NODE *list;
	int n;

	/* get n and the list */
	if ((n = xlmatch(INT, &args)->n_int) < 0) xlfail("invalid argument");
	if ((list = xlmatch(LIST, &args)) == NULL) xlfail("invalid argument");
	xllastarg(args);

	/* find the nth element */
	for (; n > 0; n--) {
		list = list->n_listnext;
		if (list == NULL || list->n_type != LIST) xlfail("invalid argument");
	}

	/* return the list beginning at the nth element */
	return (cdrflag ? list : list->n_listvalue);
}

/* xnth - return the nth element of a list */
NODE *xnth(args)
NODE *args;
{
	return (nth(args, FALSE));
}

/* xnthcdr - return the nth cdr of a list */
NODE *xnthcdr(args)
NODE *args;
{
	return (nth(args, TRUE));
}

/* xlength - return the length of a list */
NODE *xlength(args)
NODE *args;
{
	NODE *list, *val;
	int n;

	/* get the list */
	list = xlmatch(LIST, &args);
	xllastarg(args);

	/* find the length */
	for (n = 0; list != NULL; n++) list = list->n_listnext;

	/* create the value node */
	val = newnode(INT);
	val->n_int = n;

	/* return the length */
	return (val);
}

/* map - internal mapping function */
LOCAL NODE *map(args, carflag)
NODE *args;
int carflag;
{
	NODE *oldstk, fcn, lists, arglist, val, *last = NULL, *p, *x, *y;

	/* create a new stack frame */
	oldstk = xlsave(&fcn, &lists, &arglist, &val, NULL);

	/* get the function to apply */
	fcn.n_ptr = xlarg(&args);

	/* make sure there is at least one argument list */
	if (args == NULL) xlfail("too few arguments");

	/* get the argument lists */
	while (args) {
		p = newnode(LIST);
		p->n_listnext = lists.n_ptr;
		lists.n_ptr = p;
		p->n_listvalue = xlmatch(LIST, &args);
	}

	/* if the function is a symbol, get its value */
	if (fcn.n_ptr && fcn.n_ptr->n_type == SYM) fcn.n_ptr = xleval(fcn.n_ptr);

	/* loop through each of the argument lists */
	for (;;) {

		/* build an argument list from the sublists */
		arglist.n_ptr = NULL;
		for (x = lists.n_ptr; x && (y = x->n_listvalue); x = x->n_listnext) {
			p = newnode(LIST);
			p->n_listnext = arglist.n_ptr;
			arglist.n_ptr = p;
			p->n_listvalue = (carflag ? y->n_listvalue : y);
			x->n_listvalue = y->n_listnext;
		}

		/* quit if any of the lists were empty */
		if (x) break;

		/* apply the function to the arguments */
		p = newnode(LIST);

		if (val.n_ptr) {
			last->n_listnext = p;
		} else val.n_ptr = p;

		last = p;
		p->n_listvalue = xlapply(fcn.n_ptr,arglist.n_ptr);
	}

	/* restore the previous stack frame */
	xlstack = oldstk;

	/* return the last test expression value */
	return (val.n_ptr);
}

/* xmapcar - builtin function 'mapcar' */
NODE *xmapcar(args)
NODE *args;
{
	return (map(args, TRUE));
}

/* xmaplist - builtin function 'maplist' */
NODE *xmaplist(args)
NODE *args;
{
	return (map(args, FALSE));
}

/* xrplca - replace the car of a list node */
NODE *xrplca(args)
NODE *args;
{
	NODE *list, *newcar;

	/* get the list and the new car */
	if ((list = xlmatch(LIST, &args)) == NULL) xlfail("null list");
	newcar = xlarg(&args);
	xllastarg(args);

	/* replace the car */
	list->n_listvalue = newcar;

	/* return the list node that was modified */
	return (list);
}

/* xrplcd - replace the cdr of a list node */
NODE *xrplcd(args)
NODE *args;
{
	NODE *list, *newcdr;

	/* get the list and the new cdr */
	if ((list = xlmatch(LIST, &args)) == NULL) xlfail("null list");
	newcdr = xlarg(&args);
	xllastarg(args);

	/* replace the cdr */
	list->n_listnext = newcdr;

	/* return the list node that was modified */
	return (list);
}

/* xnconc - destructively append lists */
NODE *xnconc(args)
NODE *args;
{
	NODE *list, *last = NULL, *val;

	/* concatenate each argument */
	for (val = NULL; args; ) {

		/* concatenate this list */
		if ((list = xlmatch(LIST, &args))) {

			/* check for this being the first non-empty list */
			if (val) {
				last->n_listnext = list;
			} else val = list;

			/* find the end of the list */
			while (list && list->n_type == LIST && list->n_listnext) list = list->n_listnext;

			/* make sure the list ended correctly */
			if (list == NULL || list->n_type != LIST) xlfail("bad list");

			/* save the new last element */
			last = list;
		}
	}

	/* return the list */
	return (val);
}

/* delete - internal delete function */
LOCAL NODE *delete(args, fcn)
NODE *args;
int (*fcn)();
{
	NODE *x, *list, *last, *val;

	/* get the expression to delete and the list */
	x = xlarg(&args);
	list = xlmatch(LIST, &args);
	xllastarg(args);

	/* delete leading matches */
	while (list && list->n_type == LIST) {
		if (!(*fcn)(x,list->n_listvalue)) break;
		list = list->n_listnext;
	}
	val = last = list;

	/* delete embedded matches */
	if (list && list->n_type == LIST) {

		/* skip the first non-matching element */
		list = list->n_listnext;

		/* look for embedded matches */
		while (list && list->n_type == LIST) {

			/* check to see if this element should be deleted */
			if ((*fcn)(x,list->n_listvalue)) {
				last->n_listnext = list->n_listnext;
			} else last = list;

			/* move to the next element */
			list = list->n_listnext;
	 	}
	}

	/* make sure the list ended in a nil */
	if (list != NULL) xlfail("bad list");

	/* return the updated list */
	return (val);
}

/* xdelete - builtin function 'delete' */
NODE *xdelete(args)
NODE *args;
{
	return (delete(args, equal));
}

/* xdelq - builtin function 'delq' */
NODE *xdelq(args)
NODE *args;
{
	return (delete(args, eq));
}

/* xatom - is this an atom? */
NODE *xatom(args)
NODE *args;
{
	NODE *arg;
	return ((arg = xlarg(&args)) == NULL || arg->n_type != LIST ? true : ((NODE *) NULL));
}

/* xsymbolp - is this an symbol? */
NODE *xsymbolp(args)
NODE *args;
{
	NODE *arg;
	return ((arg = xlarg(&args)) && arg->n_type == SYM ? true : ((NODE *) NULL));
}

/* xnumberp - is this an number? */
NODE *xnumberp(args)
NODE *args;
{
	NODE *arg;
	return ((arg = xlarg(&args)) && arg->n_type == INT ? true : ((NODE *) NULL));
}

/* xboundp - is this a value bound to this symbol? */
NODE *xboundp(args)
NODE *args;
{
	NODE *sym;
	sym = xlmatch(SYM, &args);
	return (sym->n_symvalue == s_unbound ? ((NODE *) NULL) : true);
}

/* xnull - is this null? */
NODE *xnull(args)
NODE *args;
{
	return (xlarg(&args) == NULL ? true : ((NODE *) NULL));
}

/* xlistp - is this a list? */
NODE *xlistp(args)
NODE *args;
{
	NODE *arg;
	return ((arg = xlarg(&args)) == NULL || arg->n_type == LIST ? true : ((NODE *) NULL));
}

/* xconsp - is this a cons? */
NODE *xconsp(args)
NODE *args;
{
	NODE *arg;
	return ((arg = xlarg(&args)) != NULL && arg->n_type == LIST ? true : ((NODE *) NULL));
}

/* xeq - are these equal? */
NODE *xeq(args)
NODE *args;
{
	NODE *arg1, *arg2;

	/* get the two arguments */
	arg1 = xlarg(&args);
	arg2 = xlarg(&args);
	xllastarg(args);

	/* compare the arguments */
	return (eq(arg1,arg2) ? true : ((NODE *) NULL));
}

/* xequal - are these equal? */
NODE *xequal(args)
NODE *args;
{
	NODE *arg1, *arg2;

	/* get the two arguments */
	arg1 = xlarg(&args);
	arg2 = xlarg(&args);
	xllastarg(args);

	/* compare the arguments */
	return (equal(arg1, arg2) ? true : ((NODE *) NULL));
}

