/* xlmath - xlisp builtin arithmetic functions */
#include "xlset.h"
#include "xlmath.h"
#include "xldmem.h"
#include "xlisp.h"
#include "xlsubr.h"
#include "xlsym.h"
#include "xlsys.h"
#include "xlinit.h"

/* binary - handle binary operations */
LOCAL NODE *binary(args, funct)
NODE *args;
int (*funct)();
{
	int first, ival, iarg;
	NODE *val;

	/* initialize */
	first = TRUE;
	ival = 0;

	/* evaluate and sum each argument */
	while (args != NULL) {

		/* get the next argument */
		iarg = xlmatch(INT, &args)->n_int;

		/* accumulate the result value */
		if (first) {
			ival = iarg;
			first = FALSE;
		} else ival = (*funct)(ival,iarg);
	}

	/* initialize value */
	val = newnode(INT);
	val->n_int = ival;

	/* return the result value */
	return (val);
}

/* xadd - builtin function for addition */
LOCAL int add(val, arg)
int val;
int arg;
{
	return (val + arg);
}

NODE *xadd(args)
NODE *args;
{
	return (binary(args, add));
}

/* xsub - builtin function for subtraction */
LOCAL int sub(val, arg)
int val;
int arg;
{
	return (val - arg);
}

NODE *xsub(args)
NODE *args;
{
	return (binary(args, sub));
}

/* xmul - builtin function for multiplication */
LOCAL int mul(val, arg)
int val;
int arg;
{
	return (val * arg);
}

NODE *xmul(args)
NODE *args;
{
	return (binary(args, mul));
}

/* xdiv - builtin function for division */
LOCAL int pdiv(val, arg)
int val;
int arg;
{
	return (val / arg);
}

NODE *xdiv(args)
NODE *args;
{
	return (binary(args, pdiv));
}

/* xrem - builtin function for remainder */
LOCAL int rem(val, arg)
int val;
int arg;
{
	return (val % arg);
}

NODE *xrem(args)
NODE *args;
{
	return (binary(args, rem));
}

/* xmin - builtin function for minimum */
LOCAL int min(val, arg)
int val;
int arg;
{
	return (val < arg ? val : arg);
}

NODE *xmin(args)
NODE *args;
{
	return (binary(args, min));
}

/* xmax - builtin function for maximum */
LOCAL int max(val, arg)
int val;
int arg;
{
	return (val > arg ? val : arg);
}

NODE *xmax(args)
NODE *args;
{
	return (binary(args, max));
}

/* xbitand - builtin function for bitwise and */
LOCAL int bitand(val, arg)
int val;
int arg;
{
	return (val & arg);
}

NODE *xbitand(args)
NODE *args;
{
	return (binary(args, bitand));
}

/* xbitior - builtin function for bitwise inclusive or */
LOCAL int bitior(val, arg)
int val;
int arg;
{
	return (val | arg);
}

NODE *xbitior(args)
NODE *args;
{
	return (binary(args, bitior));
}

/* xbitxor - builtin function for bitwise exclusive or */
LOCAL int bitxor(val, arg)
int val;
int arg;
{
	return (val ^ arg);
}

NODE *xbitxor(args)
NODE *args;
{
	return (binary(args, bitxor));
}

/* xbitnot - bitwise not */
LOCAL int bitnot(arg)
int arg;
{
	return (~arg);
}

/* unary - handle unary operations */
LOCAL NODE *unary(args, fcn)
NODE *args; 
int (*fcn)();
{
	NODE *rval;
	int val;

	/* evaluate the argument */
	val = xlmatch(INT, &args)->n_int;

	/* make sure there aren't any more arguments */
	xllastarg(args);

	/* convert and check the value  */
	rval = newnode(INT);
	rval->n_int = (*fcn)(val);

	/* return the result value */
	return (rval);
}

NODE *xbitnot(args)
NODE *args;
{
	return (unary(args, bitnot));
}

/* xabs - builtin function for absolute value */
LOCAL int pabs(arg)
int arg;
{
	return (arg >= 0 ? arg : -arg);
}

NODE *xabs(args)
NODE *args;
{
	return (unary(args, pabs));
}

/* xadd1 - builtin function for adding one */
LOCAL int add1(arg)
int arg;
{
	return (arg + 1);
}

NODE *xadd1(args)
NODE *args;
{
	return (unary(args, add1));
}

/* xsub1 - builtin function for subtracting one */
LOCAL int sub1(arg)
int arg;
{
	return (arg - 1);
}

NODE *xsub1(args)
NODE *args;
{
	return (unary(args, sub1));
}

/* xminus - negate a value */
LOCAL int minus(arg)
int arg;
{
	return (-arg);
}

NODE *xminus(args)
NODE *args;
{
	return (unary(args, minus));
}

/* xlss - builtin function for < */
LOCAL int lss(cmp)
int cmp;
{
	return (cmp < 0);
}

/* gettype - return the type of an argument */
LOCAL int gettype(arg)
NODE *arg;
{
	if (arg == NULL) return (LIST);
	return (arg->n_type);
}

/* compare - common compare function */
LOCAL NODE *compare(args, funct)
NODE *args;
int (*funct)();
{
	NODE *arg1, *arg2;
	int type1, type2, cmp;

	/* get argument 1 */
	arg1 = xlarg(&args);
	type1 = gettype(arg1);

	/* get argument 2 */
	arg2 = xlarg(&args);
	type2 = gettype(arg2);

	/* make sure there aren't any more arguments */
	xllastarg(args);

	/* do the compare */
	if (type1 == STR && type2 == STR)
		cmp = strcmp(arg1->n_str, arg2->n_str);
	else if (type1 == INT && type2 == INT)
		cmp = arg1->n_int - arg2->n_int;
	else
		cmp = arg1 - arg2;

	/* return result of the compare */
	if ((*funct)(cmp)) return (true);
	
	return (NULL);
}

NODE *xlss(args)
NODE *args;
{
	return (compare(args, lss));
}

/* xleq - builtin function for <= */
LOCAL int leq(cmp)
int cmp;
{
	return (cmp <= 0);
}

NODE *xleq(args)
NODE *args;
{
	return (compare(args, leq));
}

/* eql - builtin function for = */
LOCAL int eql(cmp)
int cmp;
{
	return (cmp == 0);
}

NODE *xeql(args)
NODE *args;
{
	return (compare(args, eql));
}

/* xneq - builtin function for /= */
LOCAL int neq(cmp)
int cmp;
{
	return (cmp != 0);
}

NODE *xneq(args)
NODE *args;
{
	return (compare(args, neq));
}

/* xgeq - builtin function for >= */
LOCAL int geq(cmp)
int cmp;
{
	return (cmp >= 0);
}

NODE *xgeq(args)
NODE *args;
{
	return (compare(args, geq));
}

/* xgtr - builtin function for > */
LOCAL int gtr(cmp)
int cmp;
{
	return (cmp > 0);
}

NODE *xgtr(args)
NODE *args;
{
	return (compare(args, gtr));
}

